"""
Program which creates and runs Influx database queries
Check the Influx query documentation for query syntax:
https://docs.influxdata.com/influxdb/v2.0/api-guide/client-libraries/python/#query-data-from-influxdb-with-python
"""

from classes.common_classes import SecretStore
from classes.consts import INFLUX_DEBUG_CONFIG_TITLE, INFLUX_QUERY_CONFIG_TITLE
from classes.influx_classes import InfluxConnector
from classes.py_functions import read_query_settings, write_results_to_csv
from classes.py_logger import create_logger
from classes.query_classes import QueryBuilder


class ParseQuery:
    """
    Contains all subfunctions which parse the query results
    """

    @staticmethod
    def parse_csv(query_result: dict) -> None:
        """
        Creating a CSV file from query results
        :param query_result: Input query result to write to CSV file
        """
        logging.info("Creating CSV file")
        try:
            write_results_to_csv(
                config_name=INFLUX_QUERY_CONFIG_TITLE, table=query_result
            )
        except IOError as err:
            logging.critical("Failed to write CSV file")
            raise err

    @staticmethod
    def parse_flux(query_result: dict) -> list:
        """
        Converts input flux string into a readable dictionary
        :param query_result: Input query result to write to readable format
        :return: Dictionary of key fields from the query
        # Dictionary format [{measurement, timestamp, field, value}, [...], [...]]
        """
        result_list = []
        for table in query_result:
            for record in table.records:
                result_list.append(
                    {
                        "_measurement": record.get_measurement(),
                        "_timestamp": record.get_time(),
                        "_field": record.get_field(),
                        "_value": record.get_value(),
                    }
                )
        return result_list

    @staticmethod
    def parse_stream(query_result: dict):
        """
        Not implemented
        :param query_result: Input stream string to read
        """
        raise NotImplementedError("parse_stream not implemented")


def execute_query(query: QueryBuilder) -> None:
    """
    Run command to be used under python interactive mode
    """
    query_mode = read_query_settings(INFLUX_QUERY_CONFIG_TITLE)
    query_result = None
    try:
        logging.info(f"Running query on in mode: {query_mode}")
        influx_connector.query_database(query_mode=query_mode, query=query)
        logging.info("Successfully ran query")
    except Exception as err:
        logging.critical(f"Failed to run query: {query}")
        raise err

    if query_mode == "csv":
        ParseQuery.parse_csv(query_result=query_result)
    elif query_mode == "flux":
        flux = ParseQuery.parse_flux(query_result=query_result)
        print(flux)
    else:
        ParseQuery.parse_stream(query_result=query_result)


def run_example() -> None:
    """
    Use example query
    """
    print("Running example query")
    # from(bucket: "Bucket")
    #     |> range(start: -5m)
    #     |> filter(fn: (r) => r["_measurement"] == "fx-1" or r["_measurement"] == "mx-1")
    query_builder = QueryBuilder(bucket="Bucket", start_range="-5m")
    query_builder.append_filter(field="_measurement", value="fx-1", joiner="or")
    query_builder.append_filter(field="_measurement", value="mx-1")
    query_builder.append_filter("_measurement", "dc-1", new_band=True)
    execute_query(query_builder)


def main() -> None:
    """
    Classes runtime which creates a query to an Influx database to view the tables
    """
    print(
        "To start run this file with 'python -i ./influx_query.py'\n"
        "Build a query using QueryBuilder() run QueryBuilder.help() for info\n"
        "***************\n"
        "Then run the following with your created query: run_int_query(query)\n"
        "Or run run_default() to run an example piece"
    )
    run_example()


if __name__ == "__main__":
    logging = create_logger(INFLUX_DEBUG_CONFIG_TITLE)
    secret_store = SecretStore(has_mqtt_access=False, has_influx_access=True)
    influx_connector = InfluxConnector(secret_store.influx_secrets)
    logging.info("Attempting health check for InfluxDB")
    try:
        influx_connector.health_check()
        logging.info("Successfully connected to InfluxDB server")
    except Exception as error:
        logging.critical("Failed to connect InfluxDB server")
        raise error
    main()



"""
Contains all functions required to setup logging
"""

import configparser
import logging
import os
from logging import Logger
from logging.handlers import RotatingFileHandler, TimedRotatingFileHandler

from classes.consts import CONFIG_FILENAME
from classes.custom_exceptions import MissingConfigurationError


class LoggingTools:
    """
    Class contains all tools required to create loggers
    """

    def __init__(self, config_name: str, logger: Logger) -> None:
        """
        Initialization of logging class
        """
        self._debug_level = None
        self._file_format = None
        self._date_format = None
        self._is_file_logging = None
        self._log_rotation = None
        self._file_location = None
        self._file_path = None
        self._max_file_bytes = None
        self._max_file_no = None
        self._read_configs(config_name)

        self._create_stdout_logger(logger=logger)
        if self._is_file_logging and self._log_rotation == "size_based":
            self._create_rotating_file_logger(logger=logger)
        elif self._is_file_logging and self._log_rotation == "time_based":
            self._create_timed_rotating_file_logger(logger=logger)

    def _read_configs(self, config_name: str) -> None:
        """
        Reads config file and parses and stores the result
        """
        config_parser = configparser.ConfigParser()
        try:
            config_parser.read(CONFIG_FILENAME)
            debug_dict = {
                "DEBUG": logging.DEBUG,
                "INFO": logging.INFO,
                "WARNING": logging.WARNING,
                "ERROR": logging.ERROR,
                "CRITICAL": logging.CRITICAL,
            }
            self._debug_level = debug_dict[
                config_parser.get(config_name, "debug_level")
            ]
            self._file_format = config_parser.get(config_name, "format")
            self._date_format = config_parser.get(config_name, "dateformat")
            self._is_file_logging = config_parser.getboolean(
                config_name, "file_logging"
            )

            if None in [self._debug_level, self._file_format, self._date_format]:
                logging.critical("Failed to read basic logger configs")
                raise MissingConfigurationError("Failed to read basic logger configs")

            if self._is_file_logging:
                self._log_rotation = config_parser.get(config_name, "log_rotation")

                self._file_location = config_parser.get(config_name, "file_location")
                file_name = config_parser.get(config_name, "file_name")
                self._file_path = self._file_location + file_name
                self._max_file_bytes = int(
                    config_parser.get(config_name, "max_file_bytes")
                )
                self._max_file_no = int(config_parser.get(config_name, "max_file_no"))

                if None in [
                    self._file_location,
                    self._file_path,
                    self._max_file_bytes,
                    self._max_file_no,
                ]:
                    logging.critical("Failed to read file logger settings in configs")
                    raise MissingConfigurationError(
                        "Failed to read file logger settings in configs"
                    )
        except Exception as err:
            logging.critical("An unexpected exception has occurred")
            raise err

    def _create_stdout_logger(self, logger: Logger) -> None:
        """
        Creates a standard STDOUT logger
        """
        logger.setLevel(self._debug_level)
        stream_handler = logging.StreamHandler()
        stream_handler.setLevel(self._debug_level)
        log_formatter = logging.Formatter(
            fmt=self._file_format, datefmt=self._date_format
        )
        stream_handler.setFormatter(log_formatter)
        logger.addHandler(stream_handler)
        logging.info("Created stdout logger")

    def _create_rotating_file_logger(self, logger: Logger) -> None:
        """
        Creates a rotating file logger which limits log files size
        and when exceeding that size, creates a new log file
        """
        if not os.path.exists(self._file_location):
            os.makedirs(self._file_location)

        rotating_handler = RotatingFileHandler(
            filename=self._file_path,
            maxBytes=self._max_file_bytes,
            backupCount=self._max_file_no,
        )
        rotating_handler.setLevel(self._debug_level)
        log_formatter = logging.Formatter(
            fmt=self._file_format, datefmt=self._date_format
        )
        rotating_handler.setFormatter(log_formatter)
        logger.addHandler(rotating_handler)
        logging.info(f"Created rotating file log file at {self._file_path}")

    def _create_timed_rotating_file_logger(self, logger: Logger) -> None:
        """
        Creates a rotating file logger which limits log files size
        and when exceeding that size, creates a new log file
        """
        if not os.path.exists(self._file_location):
            os.makedirs(self._file_location)
        rotating_time_handler = TimedRotatingFileHandler(
            filename=self._file_path,
            when="midnight",
            backupCount=self._max_file_no,
        )
        rotating_time_handler.suffix = "%Y-%m-%d"
        rotating_time_handler.setLevel(self._debug_level)
        log_formatter = logging.Formatter(
            fmt=self._file_format, datefmt=self._date_format
        )
        rotating_time_handler.setFormatter(log_formatter)
        logger.addHandler(rotating_time_handler)
        logging.info(f"Created time rotating file log file at {self._file_path}")


def create_logger(config_name: str) -> logging:
    """
    Creates a logging instance, can be customized through the config.ini
    :param config_name: Section under the config for the configuration to pull data from
    :return: Logger for logging
    """
    logger = logging.getLogger()
    LoggingTools(config_name=config_name, logger=logger)
    return logging



"""
File which contains classes required by both Influx and MQTT
This file should be minimal since both MQTT and Influx is independent.
"""


import logging
import os
from dataclasses import dataclass
from datetime import datetime

from classes.consts import MAX_PORT_RANGE
from classes.custom_exceptions import MissingCredentialsError


@dataclass
class QueuePackage:
    """
    Data class which defines values that are pushed and popped off the global stack
    """

    measurement: str = None
    time_field: datetime = None
    field: str = None


class SecretStore:
    """
    Class which reads environment secrets and stores them
    """

    def __init__(
        self, has_mqtt_access: bool = False, has_influx_access: bool = False
    ) -> None:
        """
        :param mqtt_secrests: Dictionary of secrets for MQTT server
        :param influx_secrets: Dictionary of secrets for Influx server
        """
        self._has_mqtt_access = has_mqtt_access
        self._has_influx_access = has_influx_access

        self._mqtt_secrets = None
        self._influx_secrets = None

        if self._has_mqtt_access:
            logging.info("Reading MQTT environment variables")
            self._read_env_mqtt()

        if self._has_influx_access:
            logging.info("Reading Influx environment variables")
            self._read_env_influx()

    @property
    def mqtt_secrets(self) -> dict | None:
        """
        Dictionary containing MQTT secrets
        """
        assert self._mqtt_secrets is not None, "MQTT secrets missing"
        return self._mqtt_secrets

    @property
    def influx_secrets(self) -> dict | None:
        """
        Dictionary containing Influx secrets
        """
        assert self._influx_secrets is not None, "Influx secrets missing"
        return self._influx_secrets

    def _read_env_mqtt(self) -> dict:
        """
        Gets secret details from the environment file.
        :return mqtt_store: Dictionary of secrets
        """
        try:
            mqtt_port = int(os.environ.get("MQTT_PORT"))
            if mqtt_port not in range(0, MAX_PORT_RANGE):
                raise ValueError(
                    f"MQTT port outside maximum port range, 0-{MAX_PORT_RANGE}"
                )
            self._mqtt_secrets = {
                "mqtt_host": os.environ.get("MQTT_HOST"),
                "mqtt_user": os.environ.get("MQTT_USER"),
                "mqtt_port": mqtt_port,
                "mqtt_token": os.environ.get("MQTT_TOKEN"),
                "mqtt_topic": os.environ.get("MQTT_TOPIC"),
            }
            assert "" not in self._mqtt_secrets.values()
        except Exception as err:
            logging.critical("Ran into error when reading environment variables")
            raise MissingCredentialsError(
                "Ran into error when reading environment variables"
            ) from err

    def _read_env_influx(self) -> dict:
        """
        Gets secret details from the environment file.
        :return influx_store: Dictionary of secrets
        """
        try:
            self._influx_secrets = {
                "influx_url": os.environ.get("INFLUX_URL"),
                "influx_org": os.environ.get("INFLUX_ORG"),
                "influx_bucket": os.environ.get("INFLUX_BUCKET"),
                "influx_token": os.environ.get("INFLUX_TOKEN"),
            }
            assert "" not in self._influx_secrets.values()
        except Exception as err:
            logging.critical("Ran into error when reading environment variables")
            raise MissingCredentialsError(
                "Ran into error when reading environment variables"
            ) from err



"""
File for storing constant variables
"""

from queue import Queue

# Configs
CONFIG_FILENAME = "config/config.ini"  # Py Functions
INFLUX_QUERY_CONFIG_TITLE = "query_settings"  # Influx Query
INFLUX_DEBUG_CONFIG_TITLE = "influx_debugger"  # Influx Query
SOLAR_DEBUG_CONFIG_TITLE = "solar_debugger"  # Solar Runtime

# Additional Consts
MAX_PORT_RANGE = 65535
TIME_PACKET_SIZE = 4  # Measured in bytes

# Multi-Threading Processing
# Size of queue, needs to be quite large for the volume of data
MAX_QUEUE_LENGTH = 150
# Time to wait when queue is full
QUEUE_WAIT_TIME = 1
THREADED_QUEUE = Queue(maxsize=MAX_QUEUE_LENGTH)
